name = "btclib"
__version__ = "2020.12.19"
__author__ = "The btclib developers"
__author_email__ = "devs@btclib.org"
__copyright__ = "Copyright (C) 2017-2020 The btclib developers"
__license__ = "MIT License"
